#include "nvfs.h"

static int nvfs_iomap_begin(struct inode *inode, loff_t offset, loff_t length, unsigned flags, struct iomap *iomap, struct iomap *srcmap)
{
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	uint64_t page = offset >> nvs->log2_page_size;
	void *map;
	bool new;

	map = nvfs_bmap(i_to_nmi(inode), page, flags & IOMAP_WRITE ? nvfs_zero_page : NULL, NULL, &new);
	if (unlikely(IS_ERR(map)))
		return PTR_ERR(map);

	iomap->flags = 0;
	iomap->bdev = inode->i_sb->s_bdev;
	iomap->offset = page << nvs->log2_page_size;
	iomap->dax_dev = nvs->dax_dev;

	if (!map) {
		iomap->type = IOMAP_HOLE;
		iomap->addr = IOMAP_NULL_ADDR;
		iomap->length = 1UL << nvs->log2_page_size;
	} else {
		iomap->type = IOMAP_MAPPED;
		iomap->addr = map - nvs->memory_map;
		iomap->length = 1UL << nvs->log2_page_size;
		/*iomap->flags |= IOMAP_F_MERGED;*/
	}

	if (new)
		iomap->flags |= IOMAP_F_NEW;

	return 0;
}

static int nvfs_iomap_end(struct inode *inode, loff_t offset, loff_t length, ssize_t written, unsigned flags, struct iomap *iomap)
{
	return 0;
}

const struct iomap_ops nvfs_iomap_ops = {
	.iomap_begin = nvfs_iomap_begin,
	.iomap_end = nvfs_iomap_end,
};


static vm_fault_t nvfs_fault(struct vm_fault *vmf)
{
	struct inode *inode = file_inode(vmf->vma->vm_file);
	vm_fault_t ret;

	if (vmf->flags & FAULT_FLAG_WRITE) {
		struct timespec64 now;
		sb_start_pagefault(inode->i_sb);
		now = current_time(inode);
		nvfs_update_time(inode, &now, S_CTIME | S_MTIME);
	}

	down_read(&i_to_nmi(inode)->dax_sem);

	ret = dax_iomap_fault(vmf, PE_SIZE_PTE, NULL, NULL, &nvfs_iomap_ops);

	up_read(&i_to_nmi(inode)->dax_sem);

	if (vmf->flags & FAULT_FLAG_WRITE) {
		sb_end_pagefault(inode->i_sb);
	}

	return ret;
}

static const struct vm_operations_struct nvfs_vm_ops = {
	.fault = nvfs_fault,
	.page_mkwrite = nvfs_fault,
	.pfn_mkwrite = nvfs_fault,
};

int nvfs_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct inode *inode = file->f_inode;
	file_accessed(file);
	inode_lock(inode);
	i_to_nmi(inode)->was_mmapped = true;
	inode_unlock(inode);
	vma->vm_ops = &nvfs_vm_ops;
	return 0;
}
